apiVersion: apiserver.config.k8s.io/v1
kind: AdmissionConfiguration
plugins:
- name: PodSecurity
  configuration:
    apiVersion: pod-security.admission.config.k8s.io/v1beta1
    kind: PodSecurityConfiguration
    defaults:
      enforce: {{ defaults["enforce"] }}
      enforce-version: {{ defaults["enforce-version"] }}
      audit: {{ defaults["audit"] }}
      audit-version: {{ defaults["audit-version"] }}
      warn: {{ defaults["warn"] }}
      warn-version: {{ defaults["warn-version"] }}
    exemptions:
      usernames: {{ exemptions["usernames"] }}
      runtimeClasses: {{ exemptions["runtimeClasses"] }}
      namespaces: {{ exemptions["namespaces"] }}
