    body {
        font-size: 14px;
        font-family: Arial;
    }
    table {
        position: absolute;
        width: 1000px;
        left: 50%;
        margin-left: -500px;
        margin-top: 100px;
        border:none;
        border-collapse: collapse;
        font-size: 14px;
    }
    table tr {
        height: 40px;
    }
    table > tbody > tr:nth-child(odd) {
        background: #FAFAFA;
    }
    table > thead > tr {
        height: 60px;
        opacity: .5;
    }
    table > tbody > tr > td:nth-child(2) > div, #stats > div {
        height: 26px;
        color: #fff;
        text-align: center;
        line-height: 27px;
        text-transform: uppercase;
        font-size: 10px;
        border-radius: 8px;
        padding: 0 1px;
        margin: 0 10px 0 0;
    }
    table > tbody > tr.ok > td:nth-child(2) > div, #stats > div.succeeded{
        background: #4CAF50;
    }
    table > tbody > tr.fail > td:nth-child(2) > div, #stats > div.failed {
        background: #f44336;
    }
    table > tbody > tr.warning > td:nth-child(2) > div, #stats > div.warned {
        background: #FF9800;
    }
    table > tbody > tr.exception > td:nth-child(2) > div, #stats > div.excepted {
        background: #607D8B;
    }
    h1 {
        width: 1000px;
        position: absolute;
        left: 50%;
        margin-left: -500px;
        margin-top: 50px;
    }
    #date {
        position: absolute;
        margin-top: 30px;
        left: 50%;
        width: 1000px;
        margin-left: -500px;
    }
    #stats {
        position: absolute;
        margin-top: 50px;
        margin-left: -500px;
        left: 50%;
        width: 1000px;
    }
    #stats > div {
        float: right;
        padding: 2px 10px;
        margin: 0 0 0 10px;
    }
